from aiohttp import web
import asyncio
import concurrent.futures
import logging
import time
import os

import paths
from .page_former import PageFormer


PATHS_WITH_FILE_UPLOAD = {'accept_update_from_file' : paths.UPDATE_FILE_PATH,\
                          'upload_plugin' :  paths.PLUGIN_INSTALL_FILE_PATH,\
                          'upload_to_storage' : None}


async def generic(request):
    logger = logging.getLogger('app.generic_web_handler')
    pf_args = []
    if request.method == 'GET':
        logger.debug("GET %s" % request.path)
        method_name = 'do_GET'
    elif request.method == 'POST':
        logger.debug("POST %s" % request.path)
        method_name = 'do_POST'
        post_data = await request.post()
        pf_args.append(post_data)
    else:
        raise web.HTTPForbidden('Method not allowed')
    return await form_page_and_respond(request, method_name, pf_args)


async def multipart_upload_handler(request):
    logger = logging.getLogger('app.multipart_upload_handler')
    error = "Unknown error"
    if request.remote == request.host.split(':')[0]:
        error = None
    else:
        if request.secure:
            cookie = request.cookies.get(PageFormer.COOKIE_NAME_HTTPS)
        else:
            cookie = request.cookies.get(PageFormer.COOKIE_NAME)
        if cookie in request.app['cookie_tokens']:
            error = None
        else:
            error = "Error: uploads allowed only with valid token cookie"
    file_path = PATHS_WITH_FILE_UPLOAD.get(request.path)
    if not error:
        reader = await request.multipart()
        while True:
            if request.app['3dp_app'].stop_flag:
                raise web.HTTPFound("/")
            field = await reader.next()
            if field and field.name == 'file':
                break
        if not file_path:
            file_path = os.path.join(paths.STORAGE_FOLDER, field.filename)
        # You cannot rely on Content-Length if transfer is chunked.
        size = 0
        with open(file_path, 'wb') as f:
            while True:
                if request.app['3dp_app'].stop_flag:
                    break
                chunk = await field.read_chunk()  # 8192 bytes by default.
                if not chunk:
                    break
                size += len(chunk)
                f.write(chunk)
        if request.app['3dp_app'].stop_flag:
            os.remove(file_path)
            raise web.HTTPFound("/")
        logger.info('Received upload size: %dB' % size)
    pf_args = [{"multipart_error" : error}]
    return await form_page_and_respond(request, 'do_POST', pf_args)


async def form_page_and_respond(request, method_name, pf_args=[], no_respose=False):
    logger = logging.getLogger('app.generic_web_handler')
    with concurrent.futures.ThreadPoolExecutor() as pool:
        pf = PageFormer(request.app['3dp_app'],
                                     request.app['3dp_content'],
                                     request.app['3dp_config'],
                                     request.host,
                                     request.path,
                                     request.secure,
                                     request.remote,
                                     request.cookies,
                                     request.app['cookie_tokens'])
        try:
            result = await asyncio.get_running_loop().run_in_executor(pool, getattr(pf, method_name), *pf_args)
        except Exception as e:
            logger.error(e) # check why some error are not logged by log.log_exception
    new_cookie_token = getattr(pf, 'new_cookie_token', None)
    if new_cookie_token:
        request.app['cookie_tokens'].append(new_cookie_token)
    if not no_respose:
        return web.Response(status=pf.status, headers=pf.headers, body=pf.body)
